import copy
import json

import pytest

from tuxrun.tuxmake import InvalidTuxBuild, TuxBuildBuild, TuxMakeBuild

metadata = {
    "results": {
        "artifacts": {"kernel": ["bzImage"], "modules": ["modules.tar.xz"]},
    },
    "build": {"target_arch": "arm64"},
}


def build_directory(directory, metadata=metadata):
    directory.mkdir()
    (directory / "metadata.json").write_text(json.dumps(metadata))
    return directory


@pytest.fixture
def directory(tmp_path):
    return build_directory(tmp_path / "tuxmake-build")


@pytest.fixture
def directory_with_invalid_metadata(tmp_path):
    d = tmp_path / "invalid-metadata"
    d.mkdir()
    (d / "metadata.json").touch()
    return d


@pytest.fixture
def directory_with_empty_metadata(tmp_path):
    d = tmp_path / "invalid-metadata"
    d.mkdir()
    (d / "metadata.json").write_text("{}", encoding="utf-8")
    return d


@pytest.fixture
def tuxmake_build(directory):
    return TuxMakeBuild(directory)


class TestTuxMakeBuild:
    def test_kernel(self, tuxmake_build, directory):
        assert tuxmake_build.kernel == f"file://{directory / 'bzImage'}"

    def test_modules(self, tuxmake_build, directory):
        assert tuxmake_build.modules[0] == f"file://{directory / 'modules.tar.xz'}"

    def test_target_arch(self, tuxmake_build):
        assert tuxmake_build.target_arch == "arm64"

    def test_no_kernel(self, tmp_path):
        metadata1 = copy.deepcopy(metadata)
        del metadata1["results"]["artifacts"]["kernel"]
        directory = build_directory(tmp_path / "build", metadata1)
        with pytest.raises(TuxMakeBuild.Invalid):
            TuxMakeBuild(directory)

    def test_no_modules(self, tmp_path):
        metadata1 = copy.deepcopy(metadata)
        del metadata1["results"]["artifacts"]["modules"]
        directory = build_directory(tmp_path / "build", metadata1)
        tuxmake_build = TuxMakeBuild(directory)
        assert not tuxmake_build.modules

    def test_no_metadata(self, tmp_path):
        with pytest.raises(InvalidTuxBuild):
            TuxMakeBuild(tmp_path)

    def test_no_directory(self, tmp_path):
        f = tmp_path / "somefile"
        f.touch()
        with pytest.raises(InvalidTuxBuild):
            TuxMakeBuild(f)

    def test_invalid_metadata(self, directory_with_invalid_metadata):
        with pytest.raises(InvalidTuxBuild):
            TuxMakeBuild(directory_with_invalid_metadata)

    def test_empty_metadata(self, directory_with_empty_metadata):
        with pytest.raises(InvalidTuxBuild):
            TuxMakeBuild(directory_with_empty_metadata)


@pytest.fixture
def url(get, mocker):
    get.side_effect = [mocker.Mock(status_code=200, text=json.dumps(metadata))]
    return "https://example.com"


@pytest.fixture
def tuxbuild_build(url):
    return TuxBuildBuild(url)


class TestTuxBuildBuild:
    def test_kernel(self, tuxbuild_build):
        assert tuxbuild_build.kernel == "https://example.com/bzImage"

    def test_modules(self, tuxbuild_build):
        assert tuxbuild_build.modules[0] == "https://example.com/modules.tar.xz"

    def test_http_error(self, get, mocker, url):
        get.side_effect = [mocker.Mock(status_code=404)]
        with pytest.raises(InvalidTuxBuild):
            TuxBuildBuild(url)



# vim: set ts=4
#
# Copyright 2021-present Linaro Limited
#
# SPDX-License-Identifier: MIT

import fnmatch
from typing import List

from tuxrun import templates
from tuxrun.exceptions import InvalidArgument


def subclasses(cls):
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in subclasses(c)]
    )


class Test:
    devices: List[str] = []
    name: str = ""
    timeout: int = 0
    need_test_definition: bool = False

    def __init__(self, timeout):
        if timeout:
            self.timeout = timeout

    @classmethod
    def select(cls, name):
        for subclass in subclasses(cls):
            if subclass.name == name:
                return subclass
        raise InvalidArgument(f"Unknown test {name}")

    @classmethod
    def list(cls, device=None):
        if device is None:
            return sorted(s.name for s in subclasses(cls) if s.name)
        return sorted(
            t.name
            for t in subclasses(cls)
            if t.name and any([fnmatch.fnmatch(device, pat) for pat in t.devices])
        )

    def validate(self, device, **kwargs):
        if not any([fnmatch.fnmatch(device.name, pat) for pat in self.devices]):
            raise InvalidArgument(
                f"Test '{self.name}' not supported on device '{device.name}'"
            )

    def _render(self, filename, **kwargs):
        return templates.tests().get_template(filename).render(**kwargs)


import tuxrun.tests.commands  # noqa: E402
import tuxrun.tests.kselftest  # noqa: E402
import tuxrun.tests.kunit  # noqa: E402
import tuxrun.tests.kvmunittests  # noqa: E402
import tuxrun.tests.libgpiod  # noqa: E402
import tuxrun.tests.libhugetlbfs  # noqa: E402
import tuxrun.tests.ltp  # noqa: E402
import tuxrun.tests.mmtests  # noqa: E402,F401
import tuxrun.tests.morello  # noqa: E402,F401
import tuxrun.tests.perf  # noqa: E402,F401
import tuxrun.tests.rcutorture  # noqa: E402,F401
import tuxrun.tests.v4l2  # noqa: E402,F401
import tuxrun.tests.vdso  # noqa: E402,F401



import json
import os

import pytest
import yaml

import tuxrun.__main__
from tuxrun.__main__ import main, start


def touch(directory, name):
    f = directory / name
    f.touch()
    return f


@pytest.fixture
def artefacts(tmp_path):
    os.chdir(tmp_path)
    touch(tmp_path, "arm.dtb")
    touch(tmp_path, "device.yaml")
    touch(tmp_path, "definition.yaml")
    touch(tmp_path, "bios.bin")
    touch(tmp_path, "bzImage")
    touch(tmp_path, "stuff.tar.gz")
    touch(tmp_path, "morestuff.tar.gz")
    touch(tmp_path, "fvp.bin")
    touch(tmp_path, "foo.tar.xz")
    touch(tmp_path, "modules.tar")
    return tmp_path


@pytest.fixture
def run(mocker):
    return mocker.patch("tuxrun.__main__.run")


@pytest.fixture
def tuxrun_args(monkeypatch):
    args = ["tuxrun", "--device", "qemu-armv5"]
    monkeypatch.setattr("sys.argv", args)
    return args


@pytest.fixture
def tuxrun_args_generate(monkeypatch):
    args = [
        "tuxrun",
        "--device",
        "qemu-i386",
        "--kernel",
        "https://storage.tuxboot.com/buildroot/i386/bzImage",
    ]
    monkeypatch.setattr("sys.argv", args)
    return args


@pytest.fixture
def lava_run_call(mocker):
    return mocker.patch("subprocess.Popen")


@pytest.fixture
def lava_run(lava_run_call, mocker):
    mocker.patch("tuxrun.results.Results.ret", return_value=0)
    proc = lava_run_call.return_value
    proc.wait.return_value = 0
    proc.communicate.return_value = (mocker.MagicMock(), mocker.MagicMock())
    return proc


def test_start_calls_main(monkeypatch, mocker):
    monkeypatch.setattr(tuxrun.__main__, "__name__", "__main__")
    main = mocker.patch("tuxrun.__main__.main")
    with pytest.raises(SystemExit):
        start()
    main.assert_called()


def test_main_usage(monkeypatch, capsys, run):
    monkeypatch.setattr("tuxrun.__main__.sys.argv", ["tuxrun"])
    with pytest.raises(SystemExit) as exc:
        main()
    assert exc.value.code == 2
    _, err = capsys.readouterr()
    assert "usage: tuxrun" in err


def test_almost_real_run(monkeypatch, tuxrun_args, lava_run, capsys):
    lava_run.stderr = [
        '{"lvl": "info", "msg": "Hello, world", "dt": "2021-04-08T18:42:25.139513"}\n'
    ]
    exitcode = main()
    assert exitcode == 0
    stdout, _ = capsys.readouterr()
    assert "Hello, world" in stdout

    monkeypatch.setattr(
        "sys.argv",
        [
            "tuxrun",
            "--device=qemu-x86_64",
            "--modules",
            "foo.tar.xz",
            "/usr/",
            "argh",
        ],
    )
    with pytest.raises(SystemExit):
        main()

    monkeypatch.setattr(
        "sys.argv",
        [
            "tuxrun",
            "--device=qemu-x86_64",
            "--overlay",
            "foo.tar.xz",
            "/usr/",
            "argh",
        ],
    )
    with pytest.raises(SystemExit):
        main()

    monkeypatch.setattr(
        "sys.argv",
        [
            "tuxrun",
            "--device=qemu-x86_64",
            "--shared",
            "foo.tar.xz",
            "/usr/",
            "argh",
        ],
    )
    with pytest.raises(SystemExit):
        main()

    monkeypatch.setattr(
        "sys.argv",
        [
            "tuxrun",
            "--device=fvp-aemva",
            "--shared",
        ],
    )
    with pytest.raises(SystemExit):
        main()


FVP_MORELLO_ARGS = [
    "--ap-romfw",
    "fvp.bin",
    "--mcp-fw",
    "fvp.bin",
    "--mcp-romfw",
    "fvp.bin",
    "--rootfs",
    "fvp.bin",
    "--scp-fw",
    "fvp.bin",
    "--scp-romfw",
    "fvp.bin",
    "--fip",
    "fvp.bin",
]


@pytest.mark.parametrize(
    "argv",
    [
        [],
        ["--device", "qemu-armv7", "--boot-args", 'bla"bl'],
        ["--device", "fvp-aemva", "--boot-args", 'bla"bl'],
        ["--device", "qemu-armv7", "--prompt", 'bla"bl'],
        ["--device", "fvp-aemva", "--prompt", 'bla"bl'],
        ["--device", "qemu-armv7", "--dtb", "arm.dtb"],
        ["--device", "qemu-armv7", "--tests", "kselftest-arm64"],
        ["--device", "qemu-arm64", "--bios", "bios.bin"],
        ["--device", "qemu-arm64", "--modules", "modules.tar"],
        ["--kernel", "https://storage.tuxboot.com/buildroot/i386/bzImage"],
        ["--device", "fvp-aemva", "--mcp-fw", "fvp.bin"],
        ["--device", "fvp-aemva", "--modules", "modules.tar"],
        ["--device", "fvp-morello-android", "--mcp-fw", "fvp.bin"],
        ["--device", "fvp-morello-android", "--test", "multicore"],
        [
            "--device",
            "fvp-morello-android",
            *FVP_MORELLO_ARGS,
            "--tests",
            "bionic",
            "--parameters",
            "BIONIC_TEST_TYPE=invalid",
        ],
        ["--device", "fvp-morello-android", *FVP_MORELLO_ARGS, "--tests", "lldb"],
        [
            "--device",
            "fvp-morello-busybox",
            *FVP_MORELLO_ARGS,
            "--tests",
            "libjpeg-turbo",
        ],
        ["--device", "fvp-morello-busybox", *FVP_MORELLO_ARGS, "--tests", "libpng"],
        ["--device", "fvp-morello-busybox", *FVP_MORELLO_ARGS, "--tests", "libpdfium"],
        ["--device", "fvp-morello-busybox", *FVP_MORELLO_ARGS, "--tests", "zlib"],
        ["--device", "fvp-morello-busybox", *FVP_MORELLO_ARGS, "--tests", "boringssl"],
        [
            "--device",
            "fvp-morello-busybox",
            *FVP_MORELLO_ARGS,
            "--kernel",
            "https://storage.tuxboot.com/buildroot/i386/bzImage",
        ],
        ["--device", "fvp-morello-ubuntu", *FVP_MORELLO_ARGS],
        [
            "--device",
            "fvp-morello-ubuntu",
            "--ap-romfw",
            "fvp.bin",
            "--mcp-fw",
            "fvp.bin",
            "--mcp-romfw",
            "fvp.bin",
            "--scp-fw",
            "fvp.bin",
            "--scp-romfw",
            "fvp.bin",
            "--fip",
            "fvp.bin",
            "--tests",
            "lldb",
        ],
        [
            "--device",
            "fvp-morello-android",
            *FVP_MORELLO_ARGS,
            "--tests",
            "lldb",
            "--parameters",
            "LLDB_URL=http://example.com/lldb.tar.xz",
        ],
        [
            "--device",
            "fvp-morello-android",
            *FVP_MORELLO_ARGS,
            "--tests",
            "libpng",
            "--parameters",
            "SYSTEM_URL=http://example.com/system.tar.xz",
        ],
        [
            "--device",
            "fvp-morello-android",
            *FVP_MORELLO_ARGS,
            "--tests",
            "libjpeg-turbo",
            "--parameters",
            "SYSTEM_URL=http://example.com/system.tar.xz",
        ],
        [
            "--device",
            "fvp-morello-android",
            *FVP_MORELLO_ARGS,
            "--tests",
            "libpdfium",
            "--parameters",
            "SYSTEM_URL=http://example.com/system.tar.xz",
        ],
        [
            "--device",
            "fvp-morello-android",
            *FVP_MORELLO_ARGS,
            "--tests",
            "libpdfium",
            "--parameters",
            "PDF_URL=http://example.com/pdfium-testfiles.tar.xz",
        ],
        [
            "--device",
            "fvp-morello-android",
            *FVP_MORELLO_ARGS,
            "--tests",
            "ltp-smoke",
        ],
        [
            "--device",
            "fvp-morello-android",
            *FVP_MORELLO_ARGS,
            "--tests",
            "kselftest",
        ],
        ["--device", "qemu-arm64", "--tests", "ltp-smoke", "ltp-smoke"],
    ],
)
def test_command_line_errors(argv, capsys, monkeypatch, mocker, artefacts):
    monkeypatch.setattr("tuxrun.__main__.sys.argv", ["tuxrun"] + argv)
    run = mocker.patch("tuxrun.__main__.run", return_value=0)
    with pytest.raises(SystemExit) as exc:
        main()
    assert exc.value.code == 2
    stdout, stderr = capsys.readouterr()
    assert "usage: tuxrun" in stderr
    assert "tuxrun: error:" in stderr
    run.assert_not_called()


def test_command_line_parameters(monkeypatch, mocker, artefacts):
    monkeypatch.setattr(
        "tuxrun.__main__.sys.argv",
        [
            "tuxrun",
            "--device",
            "fvp-morello-android",
            "--ap-romfw",
            "fvp.bin",
            "--mcp-fw",
            "fvp.bin",
            "--mcp-romfw",
            "fvp.bin",
            "--rootfs",
            "fvp.bin",
            "--scp-fw",
            "fvp.bin",
            "--scp-romfw",
            "fvp.bin",
            "--fip",
            "fvp.bin",
            "--parameters",
            "USERDATA=http://userdata.tar.xz",
        ],
    )
    run = mocker.patch("tuxrun.__main__.run", return_value=0)
    exitcode = main()
    assert exitcode == 0
    assert len(run.call_args.args) == 4
    print(run.call_args.parameters)
    assert run.call_args[0][0].parameters == {"USERDATA": "http://userdata.tar.xz"}


def test_almost_real_run_generate(tuxrun_args_generate, lava_run, capsys):
    lava_run.stderr = [
        '{"lvl": "info", "msg": "Hello, world", "dt": "2021-04-08T18:42:25.139513"}\n'
    ]
    exitcode = main()
    assert exitcode == 0
    stdout, _ = capsys.readouterr()
    assert "Hello, world" in stdout


def test_ignores_empty_line_from_lava_run_stdout(tuxrun_args, lava_run):
    lava_run.stderr = [
        '{"lvl": "info", "msg": "Hello, world", "dt": "2021-04-08T18:42:25.139513"}\n',
        "",
        '{"lvl": "info", "msg": "Hello, world", "dt": "2021-04-08T18:42:26.139513"}\n',
    ]
    exitcode = main()
    assert exitcode == 0


def test_ignores_empty_line_from_lava_run_logfile(tuxrun_args, lava_run, tmp_path):
    log = tmp_path / "log.yaml"
    tuxrun_args += ["--log-file-yaml", str(log)]
    lava_run.stderr = [
        '{"lvl": "info", "msg": "Hello, world", "dt": "2021-04-08T18:42:25.139513"}\n',
        "",
        '{"lvl": "info", "msg": "Hello, world", "dt": "2021-04-08T18:42:26.139513"}\n',
    ]
    exitcode = main()
    assert exitcode == 0
    logdata = yaml.safe_load(log.open())
    assert type(logdata[0]) is dict
    assert type(logdata[1]) is dict


def test_exit_status_is_0_on_success(tuxrun_args, lava_run):
    assert main() == 0


def test_exit_status_matches_results(tuxrun_args, lava_run, mocker):
    mocker.patch("tuxrun.results.Results.ret", return_value=1)
    assert main() == 1


def test_save_output(monkeypatch, tmp_path, run):
    monkeypatch.setattr(
        "sys.argv", ["tuxrun", "--device", "qemu-armv5", "--save-outputs"]
    )
    main()
    run.assert_called()
    options = run.call_args[0][0]
    assert (
        options.log_file
        == tmp_path / "home" / ".cache" / "tuxrun" / "tests" / "1" / "logs"
    )
    assert (
        options.log_file_html
        == tmp_path / "home" / ".cache" / "tuxrun" / "tests" / "1" / "logs.html"
    )
    assert (
        options.log_file_text
        == tmp_path / "home" / ".cache" / "tuxrun" / "tests" / "1" / "logs.txt"
    )
    assert (
        options.log_file_yaml
        == tmp_path / "home" / ".cache" / "tuxrun" / "tests" / "1" / "logs.yaml"
    )
    assert (
        options.metadata
        == tmp_path / "home" / ".cache" / "tuxrun" / "tests" / "1" / "metadata.json"
    )
    assert (
        options.results
        == tmp_path / "home" / ".cache" / "tuxrun" / "tests" / "1" / "results.json"
    )


def test_tuxbuild(get, monkeypatch, mocker, run):
    data = json.dumps(
        {
            "results": {
                "artifacts": {"kernel": ["bzImage"], "modules": ["modules.tar.xz"]},
            },
            "build": {"target_arch": "x86_64"},
        }
    )
    mocker.patch("tuxrun.__main__.get_rootfs", return_value="https://example.com")
    get.side_effect = [mocker.Mock(status_code=200, text=data)]
    monkeypatch.setattr("sys.argv", ["tuxrun", "--tuxbuild", "https://example.com"])

    main()
    run.assert_called()
    options = run.call_args[0][0]
    assert options.kernel == "https://example.com/bzImage"
    assert options.modules[0] == "https://example.com/modules.tar.xz"
    assert options.device.name == "qemu-x86_64"
    assert options.dtb is None


def test_tuxbuild_armv5(get, monkeypatch, mocker, run):
    data = json.dumps(
        {
            "results": {
                "artifacts": {"kernel": ["bzImage"], "modules": ["modules.tar.xz"]},
            },
            "build": {"target_arch": "arm"},
        }
    )
    mocker.patch("tuxrun.__main__.get_rootfs", return_value="https://example.com")
    get.side_effect = [mocker.Mock(status_code=200, text=data)]
    monkeypatch.setattr(
        "sys.argv",
        ["tuxrun", "--tuxbuild", "https://example.com", "--device", "qemu-armv5"],
    )

    main()
    run.assert_called()
    options = run.call_args[0][0]
    assert options.kernel == "https://example.com/bzImage"
    assert options.modules[0] == "https://example.com/modules.tar.xz"
    assert options.device.name == "qemu-armv5"
    assert options.dtb == "https://example.com/dtbs/versatile-pb.dtb"


def test_tuxmake_directory(monkeypatch, tmp_path, run):
    tuxmake_build = tmp_path / "build"
    tuxmake_build.mkdir()
    (tuxmake_build / "metadata.json").write_text(
        """
        {
            "results": {
                "artifacts": {"kernel": ["bzImage"], "modules": ["modules.tar.xz"]}
            },
            "build": {"target_arch": "x86_64"}
        }
        """
    )
    monkeypatch.setattr("sys.argv", ["tuxrun", "--tuxmake", str(tuxmake_build)])

    main()
    run.assert_called()
    options = run.call_args[0][0]
    assert options.kernel == f"file://{tuxmake_build}/bzImage"
    assert options.modules[0] == f"file://{tuxmake_build}/modules.tar.xz"
    assert options.device.name == "qemu-x86_64"
    assert options.dtb is None


def test_tuxmake_directory_armv5(monkeypatch, tmp_path, run):
    tuxmake_build = tmp_path / "build"
    tuxmake_build.mkdir()
    (tuxmake_build / "metadata.json").write_text(
        """
        {
            "results": {
                "artifacts": {"kernel": ["zImage"], "modules": ["modules.tar.xz"]}
            },
            "build": {"target_arch": "arm"}
        }
        """
    )
    monkeypatch.setattr(
        "sys.argv",
        ["tuxrun", "--tuxmake", str(tuxmake_build), "--device", "qemu-armv5"],
    )

    (tuxmake_build / "dtbs").mkdir()
    (tuxmake_build / "dtbs" / "versatile-pb.dtb").write_text("")
    main()
    run.assert_called()
    options = run.call_args[0][0]
    assert options.kernel == f"file://{tuxmake_build}/zImage"
    assert options.modules[0] == f"file://{tuxmake_build}/modules.tar.xz"
    assert options.device.name == "qemu-armv5"
    assert options.dtb == f"file://{tuxmake_build}/dtbs/versatile-pb.dtb"


def test_no_modules(monkeypatch, tmp_path, run):
    tuxmake_build = tmp_path / "build"
    tuxmake_build.mkdir()
    (tuxmake_build / "metadata.json").write_text(
        """
        {
            "results": {
                "artifacts": {"kernel": ["bzImage"]}
            },
            "build": {"target_arch": "x86_64"}
        }
        """
    )
    monkeypatch.setattr("sys.argv", ["tuxrun", "--tuxmake", str(tuxmake_build)])

    main()
    run.assert_called()
    options = run.call_args[0][0]
    assert not options.modules


def test_invalid_tuxmake_directory(monkeypatch, tmp_path, capsys):
    monkeypatch.setattr("sys.argv", ["tuxrun", "--tuxmake", str(tmp_path)])
    with pytest.raises(SystemExit) as exit:
        main()
        assert exit.status_code != 0
    _, err = capsys.readouterr()
    assert "metadata.json" in err


def test_modules(monkeypatch, lava_run_call, lava_run, artefacts):
    monkeypatch.setattr(
        "sys.argv",
        [
            "tuxrun",
            "--kernel=bzImage",
            "--device=qemu-x86_64",
            "--modules=foo.tar.xz",
        ],
    )
    assert main() == 0
    lava_run_call.assert_called()
    args = lava_run_call.call_args[0][0]
    assert f"{artefacts}/foo.tar.xz:{artefacts}/foo.tar.xz:ro" in args


def test_shared(monkeypatch, lava_run_call, lava_run, artefacts):
    monkeypatch.setattr(
        "sys.argv",
        [
            "tuxrun",
            "--kernel=bzImage",
            "--device=qemu-x86_64",
            "--shared",
        ],
    )
    assert main() == 0
    lava_run_call.assert_called()
    args = lava_run_call.call_args[0][0]
    assert (
        f"{artefacts}/home/.cache/tuxrun/tests/1:{artefacts}/home/.cache/tuxrun/tests/1:rw"
        in args
    )

    monkeypatch.setattr(
        "sys.argv",
        [
            "tuxrun",
            "--kernel=bzImage",
            "--device=qemu-x86_64",
            "--shared",
            "/home/",
        ],
    )
    assert main() == 0
    lava_run_call.assert_called()
    args = lava_run_call.call_args[0][0]
    assert "/home/:/home/:rw" in args

    monkeypatch.setattr(
        "sys.argv",
        [
            "tuxrun",
            "--kernel=bzImage",
            "--device=qemu-x86_64",
            "--shared",
            "/home/",
            "/mnt/home",
        ],
    )
    assert main() == 0
    lava_run_call.assert_called()
    args = lava_run_call.call_args[0][0]
    assert "/home/:/home/:rw" in args


def test_overlays(monkeypatch, lava_run_call, lava_run, artefacts):
    monkeypatch.setattr(
        "sys.argv",
        [
            "tuxrun",
            "--kernel=bzImage",
            "--device=qemu-x86_64",
            "--overlay=stuff.tar.gz",
            "--overlay=morestuff.tar.gz",
        ],
    )
    assert main() == 0
    lava_run_call.assert_called()
    args = lava_run_call.call_args[0][0]
    assert f"{artefacts}/stuff.tar.gz:{artefacts}/stuff.tar.gz:ro" in args
    assert f"{artefacts}/morestuff.tar.gz:{artefacts}/morestuff.tar.gz:ro" in args


def test_custom_commands(monkeypatch, run):
    monkeypatch.setattr(
        "sys.argv",
        ["tuxrun", "--kernel=bzImage", "--device=qemu-x86_64", "cat", "/etc/hostname"],
    )
    main()
    run.assert_called()
    options = run.call_args[0][0]
    assert len(options.tests) == 1
    assert options.tests[0].name == "commands"
    assert options.commands == ["cat", "/etc/hostname"]


def test_list_devices(mocker, monkeypatch, capsys):
    monkeypatch.setattr(
        "sys.argv",
        ["tuxrun", "--list-devices"],
    )
    with pytest.raises(SystemExit):
        main()
    stdout, stderr = capsys.readouterr()
    assert stderr == ""
    assert "qemu-i386" in stdout


def test_list_tests(mocker, monkeypatch, capsys):
    monkeypatch.setattr(
        "sys.argv",
        ["tuxrun", "--list-tests"],
    )
    with pytest.raises(SystemExit):
        main()
    stdout, stderr = capsys.readouterr()
    assert stderr == ""
    assert "ltp-smoke" in stdout


def test_update_cache(mocker, monkeypatch, capsys):
    monkeypatch.setattr(
        "sys.argv",
        ["tuxrun", "--update-cache"],
    )
    with pytest.raises(SystemExit):
        main()
    stdout, stderr = capsys.readouterr()
    assert (
        stdout
        == """Updating local cache:
* Rootfs:
  * fvp-aemva
  * qemu-arm64
  * qemu-arm64be
  * qemu-armv5
  * qemu-armv7
  * qemu-armv7be
  * qemu-i386
  * qemu-m68k
  * qemu-mips32
  * qemu-mips32el
  * qemu-mips64
  * qemu-mips64el
  * qemu-ppc32
  * qemu-ppc64
  * qemu-ppc64le
  * qemu-riscv32
  * qemu-riscv64
  * qemu-s390
  * qemu-sh4
  * qemu-sparc64
  * qemu-x86_64
* Test definitions
"""
    )


def test_save_results_json(tuxrun_args, lava_run, mocker, tmp_path):
    json = tmp_path / "results.json"
    tuxrun_args += [f"--results={json}"]
    main()
    assert json.read_text().strip() == "{}"


def test_timeouts(monkeypatch, run):
    monkeypatch.setattr(
        "sys.argv",
        [
            "tuxrun",
            "--device=qemu-x86_64",
            "--tests",
            "ltp-smoke",
            "--timeouts",
            "boot=1",
            "ltp-smoke=12",
        ],
    )
    main()
    run.assert_called()
    options = run.call_args[0][0]
    assert len(options.tests) == 1
    assert options.tests[0].name == "ltp-smoke"
    assert options.tests[0].timeout == 12
    assert options.timeouts == {"boot": 1, "ltp-smoke": 12}


# To test the qemu_binary overlay we basically fake up the responses
# for all the various readelf/ldd/qemu calls on a real system


@pytest.fixture
def tuxrun_args_qemu_overlay(monkeypatch):
    args = [
        "tuxrun",
        "--qemu-binary",
        "/usr/bin/qemu-system-aarch64",
        "--device",
        "qemu-arm64",
    ]
    monkeypatch.setattr("sys.argv", args)
    return args


fake_readelf = b"""
String dump of section '.interp':
  [     0]  /lib64/ld-linux-x86-64.so.2
"""

# a real QEMU binary links considerably more libraries
fake_ldd = b"""
        linux-vdso.so.1 (0x00007ffe29e5f000)
        libz.so.1 => /lib/x86_64-linux-gnu/libz.so.1 (0x00007f743b834000)
        libpixman-1.so.0 => /lib/x86_64-linux-gnu/libpixman-1.so.0 (0x00007f743b789000)
        libepoxy.so.0 => /lib/x86_64-linux-gnu/libepoxy.so.0 (0x00007f743b65a000)
        libcapstone.so.4 => /lib/x86_64-linux-gnu/libcapstone.so.4 (0x00007f743aff9000)
        libspice-server.so.1 => /lib/x86_64-linux-gnu/libspice-server.so.1 (0x00007f743aeca000)
        libdw.so.1 => /lib/x86_64-linux-gnu/libdw.so.1 (0x00007f743ae21000)
"""

fake_firmware_paths = b"""
/usr/share/qemu
/usr/share/seabios
/usr/lib/ipxe/qemu
"""


@pytest.fixture
def overlay_subprocess_calls(mocker):
    def my_outputs(*args, **kwargs):
        if args[0] == "readelf":
            return fake_readelf
        elif args[0] == "ldd":
            return fake_ldd
        else:
            return fake_firmware_paths

    return mocker.patch("subprocess.check_output", new=my_outputs)


def test_qemu_overlay(tuxrun_args_qemu_overlay, lava_run, overlay_subprocess_calls):
    main()


@pytest.mark.parametrize(
    "args,cnt",
    [
        ("rw", 1),
        ("rw", 2),
        ("" "'" "rw systemd.log_level=warning" "'" "", 3),
        ("r'w", 4),
        ("'rw", 5),
        ("rw'", 6),
    ],
)
def test_boot_args(monkeypatch, mocker, tmpdir, args, cnt):
    monkeypatch.setattr(
        "tuxrun.__main__.sys.argv",
        ["tuxrun", "--device", "qemu-arm64", "--boot-args"] + [args],
    )
    mocker.patch("tuxrun.__main__.Runtime.select", side_effect=SystemExit)
    mocker.patch("tuxrun.assets.__download_and_cache__", side_effect=lambda a, b: a)
    mocker.patch(
        "tuxrun.__main__.get_test_definitions", return_value="file://testdef.tar.zst"
    )
    mocker.patch("tempfile.mkdtemp", return_value=tmpdir)
    mocker.patch("shutil.rmtree")
    if cnt < 4:
        with pytest.raises(SystemExit):
            main()
    else:
        # Invalid case
        with pytest.raises(Exception):
            main()



# vim: set ts=4
#
# Copyright 2021-present Linaro Limited
#
# SPDX-License-Identifier: MIT

import argparse
import re
import sys
import os
from abc import ABC, abstractmethod
from pathlib import Path
from urllib.parse import urlparse

from tuxrun import xdg, requests


class ProgressIndicator(ABC):
    @abstractmethod
    def progress(self, percent):
        """
        This method should display the current percentage to the user
        """

    @abstractmethod
    def finish(self):
        """
        This method should display to the user that the process has finished
        """

    @classmethod
    def get(cls, name: str) -> "ProgressIndicator":
        if sys.stderr.isatty():
            return TTYProgressIndicator(name)
        else:
            return NoProgressIndicator()


class NoProgressIndicator(ProgressIndicator):
    def progress(self, percent):
        pass

    def finish(self):
        pass


class TTYProgressIndicator(ProgressIndicator):
    def __init__(self, name):
        self.name = name

    def progress(self, percent: int) -> None:
        sys.stderr.write(f"\r{self.name} ... %3d%%" % percent)

    def finish(self) -> None:
        sys.stderr.write("\n")


COMPRESSIONS = {
    ".tar.xz": ("tar", "xz"),
    ".tar.gz": ("tar", "gz"),
    ".tgz": ("tar", "gz"),
    ".gz": (None, "gz"),
    ".xz": (None, "xz"),
    ".zst": (None, "zstd"),
}


def compression(path):
    for ext, ret in COMPRESSIONS.items():
        if path.endswith(ext):
            return ret
    return (None, None)


def notnone(value, fallback):
    if value is None:
        return fallback
    return value


def get_new_output_dir(cache_dir):
    base = xdg.get_cache_dir() / "tests"
    if cache_dir:
        base = Path(f"{os.path.abspath(cache_dir)}/tests")
    base.mkdir(parents=True, exist_ok=True)
    existing = [int(f.name) for f in base.glob("[0-9]*")]
    if existing:
        new = max(existing) + 1
    else:
        new = 1
    while True:
        new_dir = base / str(new)
        try:
            new_dir.mkdir()
            break
        except FileExistsError:
            new += 1
    return new_dir


def slugify(s):
    s = s.lower().strip()
    s = re.sub(r"[^\w\s-]", "", s)
    s = re.sub(r"[\s_-]+", "-", s)
    s = re.sub(r"^-+|-+$", "", s)
    return s


def pathurlnone(string):
    if string is None:
        return None
    url = urlparse(string)
    if url.scheme in ["http", "https"]:
        return string
    if url.scheme not in ["", "file"]:
        raise argparse.ArgumentTypeError(f"Invalid scheme '{url.scheme}'")

    path = Path(string if url.scheme == "" else url.path)
    if not path.exists():
        raise argparse.ArgumentTypeError(f"{path} no such file or directory")
    return f"file://{path.expanduser().resolve()}"


def pathnone(string):
    if string is None:
        return None

    path = Path(string)
    if not path.exists():
        raise argparse.ArgumentTypeError(f"{path} no such file or directory")
    return path.expanduser().resolve()


def callback(dataset=None, header=None, method=None, token=None, url=None, **kwargs):
    if dataset and header and method and token and url:
        s = requests.get_session(retries=3)
        s.headers.update({header: token})
        if method == "POST":
            s.post(url=url)
        if method == "GET":
            s.get(url=url)


def notify(notify):
    if not notify:
        return
    if "callbacks" in notify:
        for cb in notify.get("callbacks"):
            callback(**cb)



# vim: set ts=4
#
# Copyright 2021-present Linaro Limited
#
# SPDX-License-Identifier: MIT

import html
import logging
import re
import sys
from contextlib import ContextDecorator

import yaml

from tuxrun.yaml import yaml_load

COLORS = {
    "exception": "\033[1;31m",
    "error": "\033[1;31m",
    "warning": "\033[1;33m",
    "info": "\033[1;37m",
    "debug": "\033[0;37m",
    "target": "\033[32m",
    "input": "\033[0;35m",
    "feedback": "\033[0;33m",
    "results": "\033[1;34m",
    "dt": "\033[0;90m",
    "end": "\033[0m",
}
LOG = logging.getLogger("tuxrun")

HTML_HEADER = """<!DOCTYPE html>
<html lang="en">
<head>
  <title>TuxTest log</title>
  <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
  <style>
    body { background-color: black; color: white; }
    span.pass { color: green; }
    span.alert { color: red; }
    span.err { color: #FF7F7F; }
    span.debug { color: #FFFFFF; }
    span.info { color: #CCCCCC; }
    span.lavainfo { color: #B3BEE8; }
    span.warn { color: #FFA500; }
    span.timestamp { color: #AAFFAA; }
    span.feedback { color: orange; }
  </style>
</head>
<body>
<pre>
"""

HTML_FOOTER = """</pre>
</body>
</html>"""


class Writer(ContextDecorator):
    def __init__(self, log_file, html_file, text_file, yaml_file):
        self.log_file = log_file
        self.html_file = html_file
        self.text_file = text_file
        self.yaml_file = yaml_file

        self.lineno = 0
        self.kernel_level_pattern = re.compile(r"^\<([0-7])\>")
        self.kernel_log_levels = {
            "0": "alert",
            "1": "alert",
            "2": "alert",
            "3": "err",
            "4": "warn",
            "5": "info",
            "6": "info",
            "7": "debug",
        }

    def __enter__(self):
        def fopen(p):
            if str(p) == "-":
                return sys.stdout
            return p.open("w")

        if self.log_file is not None:
            self.log_file = fopen(self.log_file)
        if self.html_file is not None:
            self.html_file = fopen(self.html_file)
            self.html_file.write(HTML_HEADER)
        if self.text_file is not None:
            self.text_file = fopen(self.text_file)
        if self.yaml_file is not None:
            self.yaml_file = fopen(self.yaml_file)
        return self

    def __exit__(self, exc_type, exc, exc_tb):
        def fclose(f):
            if f is not None and f != sys.stdout:
                f.close()

        if self.html_file is not None:
            self.html_file.write(HTML_FOOTER)
        fclose(self.log_file)
        fclose(self.text_file)
        fclose(self.html_file)
        fclose(self.yaml_file)

    def write(self, line):
        line = line.rstrip("\n")
        try:
            data = yaml_load(line)
        except yaml.YAMLError:
            sys.stdout.write(line + "\n")
            return
        if not data or not isinstance(data, dict):
            sys.stdout.write(line + "\n")
            return
        if not {"dt", "lvl", "msg"}.issubset(data.keys()):
            sys.stdout.write(line + "\n")
            return

        if self.log_file is not None:
            level = data["lvl"]
            msg = data["msg"]
            ns = " "
            if level == "feedback" and "ns" in data:
                ns = f" <{COLORS['feedback']}{data['ns']}{COLORS['end']}> "
            timestamp = data["dt"].split(".")[0]

            if level == "input" and msg[-1] == "\n":
                msg = msg[0:-1] + "⏎"
            self.log_file.write(
                f"{COLORS['dt']}{timestamp}{COLORS['end']}{ns}{COLORS[level]}{msg}{COLORS['end']}\n"
            )

        if self.yaml_file is not None:
            self.yaml_file.write("- " + line + "\n")

        if data["lvl"] in ["target", "feedback"]:
            if self.text_file is not None:
                if data["lvl"] == "feedback" and "ns" in data:
                    self.text_file.write(f"<{data['ns']}> ")
                self.text_file.write(data["msg"] + "\n")

            if self.html_file is not None:
                # Build the html output
                kernel_level = self.kernel_level_pattern.match(data["msg"])

                self.html_file.write(f"""<span id="L{self.lineno}">""")
                self.html_file.write(
                    f"""<span class="timestamp">{data['dt']}</span> """
                )
                if kernel_level:
                    cls = self.kernel_log_levels[kernel_level.group(1)]
                    self.html_file.write(
                        f"""<span class="{cls}">{html.escape(data["msg"])}</span>"""
                    )
                else:
                    ns = ""
                    if data["lvl"] == "feedback" and "ns" in data:
                        ns = f"<span class=\"feedback\">{html.escape('<' + data['ns'] + '> ')}</span>"
                    self.html_file.write(ns + html.escape(data["msg"]))
                self.html_file.write("</span>\n")
                self.lineno += 1



# vim: set ts=4
#
# Copyright 2023-present Linaro Limited
#
# SPDX-License-Identifier: MIT

from tuxrun.tests import Test


class MMTests(Test):
    devices = ["qemu-arm64", "qemu-x86_64"]
    configfile: str = ""
    full_archive: bool = False
    iterations: int = 10
    timeout = 90
    need_test_definition = True

    def render(self, **kwargs):
        kwargs["name"] = self.name
        kwargs["configfile"] = self.configfile
        kwargs["full_archive"] = self.full_archive
        kwargs["iterations"] = self.iterations
        kwargs["timeout"] = self.timeout
        return self._render("mmtests.yaml.jinja2", **kwargs)


class MMTestsDbSqliteInsertSmall(MMTests):
    configfile = "configs/config-db-sqlite-insert-small"
    name = "mmtests-db-sqlite-insert-small"


class MMTestsHpcScimarkcSmall(MMTests):
    configfile = "configs/config-hpc-scimarkc-small"
    name = "mmtests-hpc-scimarkc-small"
    iterations = 20


class MMTestsBlogbench(MMTests):
    configfile = "configs/config-io-blogbench"
    name = "mmtests-io-blogbench"
    iterations = 30


class MMTestsFioRandreadAsyncRandwrite(MMTests):
    configfile = "configs/config-io-fio-randread-async-randwrite"
    name = "mmtests-io-fio-randread-async-randwrite"


class MMTestsFioRandreadAsyncSeqwrite(MMTests):
    configfile = "configs/config-io-fio-randread-async-seqwrite"
    name = "mmtests-io-fio-randread-async-seqwrite"


class MMTestsFioRandreadSyncHeavywrite(MMTests):
    configfile = "configs/config-io-fio-randread-sync-heavywrite"
    name = "mmtests-io-fio-randread-sync-heavywrite"


class MMTestsFioRandreadSyncRandwrite(MMTests):
    configfile = "configs/config-io-fio-randread-sync-randwrite"
    name = "mmtests-io-fio-randread-sync-randwrite"


class MMTestsFsmarkSmallFileStream(MMTests):
    configfile = "configs/config-io-fsmark-small-file-stream"
    name = "mmtests-io-fsmark-small-file-stream"


class MMTestsRedisBenchmarkSmall(MMTests):
    configfile = "configs/config-memdb-redis-benchmark-small"
    name = "mmtests-memdb-redis-benchmark-small"
    iterations = 20


class MMTestsRedisMemtierSmall(MMTests):
    configfile = "configs/config-memdb-redis-memtier-small"
    name = "mmtests-memdb-redis-memtier-small"
    iterations = 20


class MMTestsSchbench(MMTests):
    configfile = "configs/config-scheduler-schbench"
    name = "mmtests-scheduler-schbench"


class MMTestsSysbenchCpu(MMTests):
    configfile = "configs/config-scheduler-sysbench-cpu"
    name = "mmtests-scheduler-sysbench-cpu"


class MMTestsSysbenchThread(MMTests):
    configfile = "configs/config-scheduler-sysbench-thread"
    name = "mmtests-scheduler-sysbench-thread"


class MMTestsAim9Disk(MMTests):
    configfile = "configs/config-workload-aim9-disk"
    name = "mmtests-workload-aim9-disk"


class MMTestsCoremark(MMTests):
    configfile = "configs/config-workload-coremark"
    name = "mmtests-workload-coremark"
    iterations = 20


class MMTestsCyclictestFineHackbench(MMTests):
    configfile = "configs/config-workload-cyclictest-fine-hackbench"
    name = "mmtests-workload-cyclictest-fine-hackbench"
    iterations = 15


class MMTestsCyclictestHackbench(MMTests):
    configfile = "configs/config-workload-cyclictest-hackbench"
    name = "mmtests-workload-cyclictest-hackbench"
    iterations = 20


class MMTestsEbizzy(MMTests):
    configfile = "configs/config-workload-ebizzy"
    name = "mmtests-workload-ebizzy"


class MMTestsPmqtestHackbench(MMTests):
    configfile = "configs/config-workload-pmqtest-hackbench"
    name = "mmtests-workload-pmqtest-hackbench"


class MMTestsStressngAfAlg(MMTests):
    configfile = "configs/config-workload-stressng-af-alg"
    name = "mmtests-workload-stressng-af-alg"


class MMTestsStressngBadAltstack(MMTests):
    configfile = "configs/config-workload-stressng-bad-altstack"
    name = "mmtests-workload-stressng-bad-altstack"


class MMTestsStressngClassIoParallel(MMTests):
    configfile = "configs/config-workload-stressng-class-io-parallel"
    name = "mmtests-workload-stressng-class-io-parallel"


class MMTestsStressngContext(MMTests):
    configfile = "configs/config-workload-stressng-context"
    name = "mmtests-workload-stressng-context"


class MMTestsStressngFork(MMTests):
    configfile = "configs/config-workload-stressng-fork"
    name = "mmtests-workload-stressng-fork"


class MMTestsStressngGet(MMTests):
    configfile = "configs/config-workload-stressng-get"
    name = "mmtests-workload-stressng-get"


class MMTestsStressngGetdent(MMTests):
    configfile = "configs/config-workload-stressng-getdent"
    name = "mmtests-workload-stressng-getdent"


class MMTestsStressngMadvise(MMTests):
    configfile = "configs/config-workload-stressng-madvise"
    name = "mmtests-workload-stressng-madvise"


class MMTestsStressngMmap(MMTests):
    configfile = "configs/config-workload-stressng-mmap"
    name = "mmtests-workload-stressng-mmap"


class MMTestsStressngVmSplice(MMTests):
    configfile = "configs/config-workload-stressng-vm-splice"
    name = "mmtests-workload-stressng-vm-splice"


class MMTestsStressngZombie(MMTests):
    configfile = "configs/config-workload-stressng-zombie"
    name = "mmtests-workload-stressng-zombie"


class MMTestsUsemem(MMTests):
    configfile = "configs/config-workload-usemem"
    name = "mmtests-workload-usemem"


class MMTestsScaleIoProcesses(MMTests):
    configfile = "configs/config-workload-will-it-scale-io-processes"
    name = "mmtests-workload-will-it-scale-io-processes"


class MMTestsScaleIoThreads(MMTests):
    configfile = "configs/config-workload-will-it-scale-io-threads"
    name = "mmtests-workload-will-it-scale-io-threads"


class MMTestsScalePfProcesses(MMTests):
    configfile = "configs/config-workload-will-it-scale-pf-processes"
    name = "mmtests-workload-will-it-scale-pf-processes"


class MMTestsScalePfThreads(MMTests):
    configfile = "configs/config-workload-will-it-scale-pf-threads"
    name = "mmtests-workload-will-it-scale-pf-threads"


class MMTestsScaleSysProcesses(MMTests):
    configfile = "configs/config-workload-will-it-scale-sys-processes"
    name = "mmtests-workload-will-it-scale-sys-processes"


class MMTestsScaleSysThreads(MMTests):
    configfile = "configs/config-workload-will-it-scale-sys-threads"
    name = "mmtests-workload-will-it-scale-sys-threads"



import os
import subprocess
from pathlib import Path

import pytest

from tuxrun.runtimes import DockerRuntime, NullRuntime, PodmanRuntime, Runtime


def test_select():
    assert Runtime.select("docker") == DockerRuntime
    assert Runtime.select("null") == NullRuntime
    assert Runtime.select("podman") == PodmanRuntime


def test_cmd_null():
    runtime = Runtime.select("null")()
    assert runtime.cmd(["hello", "world"]) == ["hello", "world"]

    runtime.bind("/hello/world")
    assert runtime.cmd(["hello", "world"]) == ["hello", "world"]


def test_cmd_podman():
    runtime = Runtime.select("podman")()
    runtime.name("name")
    runtime.image("image")
    args = [
        "podman",
        "run",
        "--log-driver=none",
        "--rm",
        "--hostname",
        "tuxrun",
        "-v",
        "/boot:/boot:ro",
        "-v",
        "/lib/modules:/lib/modules:ro",
    ]
    if Path("/dev/kvm").exists():
        args.extend(
            [
                "--device",
                "/dev/kvm:/dev/kvm:rw",
            ]
        )
    if Path(f"/var/tmp/.guestfs-{os.getuid()}").exists():
        args.extend(
            [
                "-v",
                f"/var/tmp/.guestfs-{os.getuid()}:/var/tmp/.guestfs-0:rw",
            ]
        )
    assert runtime.cmd(["hello", "world"]) == args + [
        "--name",
        "name",
        "image",
        "hello",
        "world",
    ]

    runtime.bind("/hello/world")
    assert runtime.cmd(["hello", "world"]) == args + [
        "-v",
        "/hello/world:/hello/world:rw",
        "--name",
        "name",
        "image",
        "hello",
        "world",
    ]


def test_kill_null(mocker):
    runtime = Runtime.select("null")()
    runtime.__proc__ = None
    runtime.kill()

    runtime.__proc__ = mocker.MagicMock()
    runtime.kill()
    runtime.__proc__.send_signal.assert_called_once_with(15)


def test_kill_podman(mocker):
    popen = mocker.patch("subprocess.Popen")
    runtime = Runtime.select("podman")()
    runtime.kill()
    popen.assert_called_once_with(
        ["podman", "stop", "--time", "60", None],
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        preexec_fn=os.setpgrp,
    )
    assert len(runtime.__sub_procs__) == 1


def test_kill_podman_raise(mocker):
    popen = mocker.patch("subprocess.Popen", side_effect=FileNotFoundError)
    runtime = Runtime.select("podman")()
    runtime.kill()
    popen.assert_called_once_with(
        ["podman", "stop", "--time", "60", None],
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        preexec_fn=os.setpgrp,
    )
    assert len(runtime.__sub_procs__) == 0


def test_pre_run_docker(tmp_path):
    runtime = Runtime.select("docker")()
    runtime.pre_run(tmp_path)
    assert runtime.__bindings__[-1] == (
        "/var/run/docker.sock",
        "/var/run/docker.sock",
        False,
        False,
    )


def test_pre_run_null():
    runtime = Runtime.select("null")()
    runtime.pre_run(None)


def test_pre_run_podman(mocker, tmp_path):
    (tmp_path / "podman.sock").touch()
    popen = mocker.patch("subprocess.Popen")

    runtime = Runtime.select("podman")()
    runtime.pre_run(tmp_path)
    assert runtime.__pre_proc__ is not None
    popen.assert_called_once()


def test_pre_run_podman_errors(mocker, tmp_path):
    popen = mocker.patch("subprocess.Popen")
    sleep = mocker.patch("time.sleep")

    runtime = Runtime.select("podman")()
    with pytest.raises(Exception) as exc:
        runtime.pre_run(tmp_path)
    assert exc.match("Unable to create podman socket at ")
    assert exc.match("podman.sock")
    popen.assert_called_once()
    sleep.assert_called()


def test_post_run_null():
    runtime = Runtime.select("null")()
    runtime.post_run()


def test_post_run_podman(mocker):
    runtime = Runtime.select("podman")()
    runtime.post_run()

    runtime.__pre_proc__ = mocker.MagicMock()
    runtime.post_run()
    runtime.__pre_proc__.kill.assert_called_once_with()
    runtime.__pre_proc__.wait.assert_called_once_with()


def test_run(mocker):
    popen = mocker.patch("subprocess.Popen")

    runtime = Runtime.select("podman")()
    runtime.name("name")
    runtime.image("image")
    with runtime.run(["hello", "world"]):
        popen.assert_called_once()
        assert runtime.__proc__ is not None
        assert runtime.__ret__ is None
    runtime.__proc__.wait.assert_called_once()


def test_run_errors(mocker, tmp_path):
    popen = mocker.patch("subprocess.Popen", side_effect=FileNotFoundError)

    runtime = Runtime.select("podman")()
    runtime.name("name")
    runtime.image("image")
    with pytest.raises(FileNotFoundError):
        with runtime.run(["hello", "world"]):
            pass
    popen.assert_called_once()

    popen = mocker.patch("subprocess.Popen", side_effect=Exception)
    runtime = Runtime.select("podman")()
    runtime.name("name")
    runtime.image("image")
    with pytest.raises(Exception):
        with runtime.run(["hello", "world"]):
            pass
    popen.assert_called_once()

    # Test duplicated source bindings
    popen = mocker.patch("subprocess.Popen", side_effect=FileNotFoundError)

    runtime = Runtime.select("podman")()
    runtime.name("name")
    runtime.image("image")
    runtime.bind("/hello", "/world")
    runtime.bind("/hello", "/world2")
    with pytest.raises(Exception) as exc:
        with runtime.run(["hello", "world"]):
            pass
    assert exc.match("Duplicated mount source '/hello'")
    popen.assert_not_called()

    # Test duplicated destination bindings
    runtime = Runtime.select("podman")()
    runtime.name("name")
    runtime.image("image")
    runtime.bind("/hello", "/world")
    runtime.bind("/hello2", "/world")
    with pytest.raises(Exception) as exc:
        with runtime.run(["hello", "world"]):
            pass
    assert exc.match("Duplicated mount destination '/world'")
    popen.assert_not_called()



from tuxrun.writer import Writer


def test_write_log_file(tmp_path):
    log_file = tmp_path / "logs"
    html_file = tmp_path / "logs.html"
    text_file = tmp_path / "logs.txt"
    yaml_file = tmp_path / "logs.yaml"
    with Writer(log_file, html_file, text_file, yaml_file) as writer:
        writer.write(
            '{"lvl": "info", "msg": "Hello, world", "dt": "2021-04-08T18:42:25.139513"}'
        )

    log_file.read_text(
        encoding="utf-8"
    ) == '{"lvl": "info", "msg": "Hello, world", "dt": "2021-04-08T18:42:25.139513"}\n'

    assert (
        html_file.read_text(encoding="utf-8")
        == """<!DOCTYPE html>
<html lang="en">
<head>
  <title>TuxTest log</title>
  <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
  <style>
    body { background-color: black; color: white; }
    span.pass { color: green; }
    span.alert { color: red; }
    span.err { color: #FF7F7F; }
    span.debug { color: #FFFFFF; }
    span.info { color: #CCCCCC; }
    span.lavainfo { color: #B3BEE8; }
    span.warn { color: #FFA500; }
    span.timestamp { color: #AAFFAA; }
    span.feedback { color: orange; }
  </style>
</head>
<body>
<pre>
</pre>
</body>
</html>"""
    )
    assert text_file.read_text(encoding="utf-8") == ""


def test_write_stdout(capsys):
    with Writer("-", None, None, None) as writer:
        writer.write(
            '{"lvl": "info", "msg": "Hello, world", "dt": "2021-04-08T18:42:25.139513"}'
        )

    out, err = capsys.readouterr()
    assert "\x1b[0m \x1b[1;37mHello, world\x1b[0m\n" in out


def test_write_stdout_logs(capsys, tmp_path):
    html_file = tmp_path / "logs.html"
    text_file = tmp_path / "logs.txt"
    yaml_file = tmp_path / "logs.yaml"
    with Writer("-", html_file, text_file, yaml_file) as writer:
        writer.write(
            '{"lvl": "info", "msg": "Hello, world", "dt": "2021-04-08T18:42:25.139513"}'
        )

    out, err = capsys.readouterr()
    assert "\x1b[0m \x1b[1;37mHello, world\x1b[0m\n" in out

    assert (
        yaml_file.read_text(encoding="utf-8")
        == """- {"lvl": "info", "msg": "Hello, world", "dt": "2021-04-08T18:42:25.139513"}\n"""
    )
    assert (
        html_file.read_text(encoding="utf-8")
        == """<!DOCTYPE html>
<html lang="en">
<head>
  <title>TuxTest log</title>
  <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
  <style>
    body { background-color: black; color: white; }
    span.pass { color: green; }
    span.alert { color: red; }
    span.err { color: #FF7F7F; }
    span.debug { color: #FFFFFF; }
    span.info { color: #CCCCCC; }
    span.lavainfo { color: #B3BEE8; }
    span.warn { color: #FFA500; }
    span.timestamp { color: #AAFFAA; }
    span.feedback { color: orange; }
  </style>
</head>
<body>
<pre>
</pre>
</body>
</html>"""
    )
    assert text_file.read_text(encoding="utf-8") == ""


def test_write_stdout_feedback(capsys):
    with Writer("-", None, None, None) as writer:
        writer.write(
            '{"lvl": "feedback", "msg": "Hello, world", "dt": "2021-04-08T18:42:25.139513", "ns": "testing"}'
        )

    out, err = capsys.readouterr()
    assert "\x1b[0m <\x1b[0;33mtesting\x1b[0m> \x1b[0;33mHello, world\x1b[0m\n" in out


def test_write_stdout_feedback_logs(capsys, tmp_path):
    html_file = tmp_path / "logs.html"
    text_file = tmp_path / "logs.txt"
    yaml_file = tmp_path / "logs.yaml"
    with Writer("-", html_file, text_file, yaml_file) as writer:
        writer.write(
            '{"lvl": "feedback", "msg": "Hello, world", "dt": "2021-04-08T18:42:25.139513", "ns": "testing"}'
        )

    out, err = capsys.readouterr()
    assert "\x1b[0m <\x1b[0;33mtesting\x1b[0m> \x1b[0;33mHello, world\x1b[0m\n" in out

    assert (
        yaml_file.read_text(encoding="utf-8")
        == """- {"lvl": "feedback", "msg": "Hello, world", "dt": "2021-04-08T18:42:25.139513", "ns": "testing"}\n"""
    )
    assert (
        html_file.read_text(encoding="utf-8")
        == """<!DOCTYPE html>
<html lang="en">
<head>
  <title>TuxTest log</title>
  <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
  <style>
    body { background-color: black; color: white; }
    span.pass { color: green; }
    span.alert { color: red; }
    span.err { color: #FF7F7F; }
    span.debug { color: #FFFFFF; }
    span.info { color: #CCCCCC; }
    span.lavainfo { color: #B3BEE8; }
    span.warn { color: #FFA500; }
    span.timestamp { color: #AAFFAA; }
    span.feedback { color: orange; }
  </style>
</head>
<body>
<pre>
<span id="L0"><span class="timestamp">2021-04-08T18:42:25.139513</span> <span class="feedback">&lt;testing&gt; </span>Hello, world</span>
</pre>
</body>
</html>"""
    )
    assert text_file.read_text(encoding="utf-8") == "<testing> Hello, world\n"


def test_writer_invalid_yaml(capsys, tmpdir):
    data = '{"lvl": "feedback", "msg": "Hello, world", "dt": "2021-04-08T18:42:25.139513", "ns": "testing"}'
    with Writer(tmpdir / "logs", None, None, tmpdir / "logs.yaml") as writer:
        writer.write("{")
        writer.write("hello world")
        writer.write("{}")
        writer.write(data)
        writer.write("{hello: world}")
    out, err = capsys.readouterr()
    assert (
        out
        == """{
hello world
{}
{hello: world}
"""
    )
    assert err == ""
    assert (tmpdir / "logs").read_text(
        encoding="utf-8"
    ) == "\x1b[0;90m2021-04-08T18:42:25\x1b[0m <\x1b[0;33mtesting\x1b[0m> \x1b[0;33mHello, world\x1b[0m\n"
    assert (tmpdir / "logs.yaml").read_text(encoding="utf-8") == f"- {data}\n"



# vim: set ts=4
#
# Copyright 2021-present Linaro Limited
#
# SPDX-License-Identifier: MIT

import logging
import re

import yaml

from tuxrun.yaml import yaml_load

LOG = logging.getLogger("tuxrun")
PATTERN = re.compile(r"^(\d+_)")


class Results:
    def __init__(self, tests, artefacts):
        self.__artefacts__ = artefacts.copy()
        # Add overlays
        for index, overlay in enumerate(self.__artefacts__.get("overlays", [])):
            self.__artefacts__[f"overlay-{index:02}"] = overlay[0]
        self.__data__ = {}
        self.__metadata__ = {}
        self.__post_processed = False
        self.__tests__ = ["lava"] + [t.name for t in tests]
        self.__ret__ = 0

    def parse(self, line):
        try:
            data = yaml_load(line)
        except yaml.YAMLError:
            LOG.debug(line)
            return
        if not data or not isinstance(data, dict):
            LOG.debug(line)
            return
        if data.get("lvl") != "results":
            return

        test = data.get("msg")
        if not {"case", "definition"}.issubset(test.keys()):
            LOG.debug(line)
            return

        definition = re.sub(PATTERN, "", test.pop("definition"))
        case = re.sub(PATTERN, "", test.pop("case"))
        # download action can be duplicated
        if (
            definition == "lava"
            and case.endswith("-download")
            and test.get("extra", {}).get("label")
        ):
            label = test["extra"]["label"]
            label = label[len("rootfs.") :] if label.startswith("rootfs.") else label
            self.__data__.setdefault(definition, {}).setdefault(case, {})[label] = test
            if label in self.__artefacts__:
                self.__data__[definition][case][label]["url"] = (
                    self.__artefacts__[label]
                    if isinstance(self.__artefacts__[label], str)
                    else None
                    if not self.__artefacts__[label]
                    else self.__artefacts__[label][0]
                )
        else:
            self.__data__.setdefault(definition, {})[case] = test
        if test["result"] == "fail":
            self.__ret__ = 1

    def __post_process(self):
        if self.__post_processed:
            return
        self.__post_processed = True

        if set(self.__tests__) != set(self.__data__.keys()):
            self.__ret__ = 2

        # Add qemu info
        if self.__data__.get("lava", {}).get("execute-qemu", {}).get("extra"):
            self.__metadata__ = {
                "arch": self.__data__["lava"]["execute-qemu"]["extra"].get("job_arch"),
                "host_arch": self.__data__["lava"]["execute-qemu"]["extra"].get(
                    "host_arch"
                ),
                "qemu_version": self.__data__["lava"]["execute-qemu"]["extra"].get(
                    "qemu_version"
                ),
            }
        # Add artefacts with url and checksum
        self.__metadata__["artefacts"] = {}
        for lava_key in ["file-download", "http-download"]:
            for k, v in self.__data__.get("lava", {}).get(lava_key, {}).items():
                if "url" in v:
                    self.__metadata__["artefacts"][k] = {
                        "url": v["url"],
                        "sha256sum": v["extra"]["sha256sum"],
                    }

        # Add test durations
        self.__metadata__["durations"] = {"tests": {}}
        for test in self.__tests__[1:]:
            self.__metadata__["durations"]["tests"][test] = (
                self.__data__.get("lava", {}).get(test, {}).get("duration", 0)
            )
        self.__metadata__["durations"]["tests"]["boot"] = (
            self.__data__.get("lava", {}).get("login-action", {}).get("duration", 0)
        )

    @property
    def data(self):
        self.__post_process()
        return self.__data__

    @property
    def metadata(self):
        self.__post_process()
        return self.__metadata__

    def ret(self):
        self.__post_process()
        return self.__ret__



# vim: set ts=4
#
# Copyright 2021-present Linaro Limited
#
# SPDX-License-Identifier: MIT

from typing import Dict, List

from tuxrun.exceptions import InvalidArgument


def subclasses(cls):
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in subclasses(c)]
    )


class Device:
    name: str = ""
    flag_use_pre_run_cmd: bool = False
    flag_cache_rootfs: bool = False

    @classmethod
    def select(cls, name):
        for subclass in subclasses(cls):
            if subclass.name == name:
                return subclass
        raise InvalidArgument(
            f"Unknown device {name}. Available: {', '.join([c.name for c in cls.list()])}"
        )

    @classmethod
    def list(cls) -> List["Device"]:
        return sorted([s for s in subclasses(cls) if s.name], key=lambda d: d.name)

    def validate(self, **kwargs):
        raise NotImplementedError  # pragma: no cover

    def default(self, options) -> None:
        raise NotImplementedError  # pragma: no cover

    def definition(self, **kwargs) -> str:
        raise NotImplementedError  # pragma: no cover

    def device_dict(self, context: Dict) -> str:
        raise NotImplementedError  # pragma: no cover

    def extra_assets(self, tmpdir, **kwargs) -> List[str]:
        return []


import tuxrun.devices.fvp  # noqa: E402
import tuxrun.devices.qemu  # noqa: E402,F401
